#include <rdp.h>
#include <stdlib.h>

struct _buffer
{
  symbol_t* symbol;
  gpointer attrib;
};

gpointer leaf_new (gpointer data)
{
  return g_node_new (data);
}

gpointer tree_new (rule_t* rule)
{
  return g_node_new (rule);
}

gpointer tree_add (gpointer tree, gpointer data)
{
  return g_node_append (tree, data);
}

void tree_delete (gpointer tree)
{
  g_node_destroy (tree);
}

static void rdp_init (GTypeInstance* instance, gpointer g_class)
{
  Rdp* self = RDP(instance);
  self->cb = NULL;
  self->data = NULL;
  self->buffer = NULL;
  self->start = NULL;
}

static void rdp_finalize (GObject* obj)
{
  RdpClass* klass;
  GObject* parent_class;
  Rdp* self;
  self = RDP(obj);
  g_free (self->start);
  klass = RDP_GET_CLASS(obj);
  parent_class = g_type_class_peek_parent (klass);
  G_OBJECT_CLASS(parent_class)->finalize (obj);
}

static void rdp_class_init (RdpClass* klass)
{
  GObjectClass* gobj_class = G_OBJECT_CLASS(klass);
  gobj_class->finalize = rdp_finalize;
}

GType rdp_get_type ()
{
  static GType type = 0;
  if (type == 0)
    {
      static const GTypeInfo info =
	{
	  sizeof (RdpClass),
	  NULL,
	  NULL,
	  (GClassInitFunc)rdp_class_init,
	  NULL,
	  NULL,
	  sizeof (Rdp),
	  0,
	  rdp_init
	};
      type = g_type_register_static (GRAMMAR_TYPE, "RdpType", &info, 0);
    }
  return type;
}

Rdp* rdp_new (nextcb cb, gpointer data, gint value)
{

  Rdp* parser;

  parser = g_object_new (RDP_TYPE, NULL);

  parser->cb = cb;
  parser->data = data;
  parser->start = symbol_new (FALSE, value);

  parser->buffer = g_list_append (NULL, NULL);

  return parser;

}

void rdp_delete (Rdp* parser)
{

  g_object_unref (parser);

}

symbol_t* buffer_next (Rdp* parser, gpointer* attrib)
{

  buffer_t* buffer;

  if (parser->buffer->next == NULL)
    {
      buffer = g_malloc (sizeof (buffer_t));
      buffer->symbol = g_malloc (sizeof (symbol_t));
      buffer->symbol->terminal = TRUE;
      buffer->symbol->value = parser->cb (parser->data, &(buffer->attrib));
      g_list_append (parser->buffer, buffer);
    }

  parser->buffer = g_list_next (parser->buffer);
  buffer = (buffer_t*) parser->buffer->data;

  if (attrib)
    *attrib = buffer->attrib;

  return buffer->symbol;
      
}

gboolean rdp_step (Rdp* parser, symbol_t* symbol, gpointer* attrib)
{

  GList* l;
  GList* buffer;
  gpointer attr;

  buffer = parser->buffer;

  if (symbol != NULL && symbol->terminal)
    {
      symbol_t* s;
      s = buffer_next (parser, &attr);
      if (!symbol_equal (symbol, s))
	{
	  parser->buffer = buffer;
	  return FALSE;
	}
      *attrib = leaf_new (attr);
      return TRUE;
    }

  for (l = grammar_get_rules (parser, symbol); l != NULL; l = g_list_next (l))
    {

      rule_t* rule;
      GList* m;

      rule = (rule_t*) l->data;


      *attrib = tree_new (symbol_copy (symbol));

      m = grammar_get_rule (rule);

      /*
      if (m == NULL)
	{
	  tree_add (*attrib, leaf_new (g_string_new ("")));
	  return TRUE;
	}
      */

      while (m != NULL)
	{

	  symbol_t* s;

	  s = (symbol_t*) m->data;

	  if (!rdp_step (parser, s, &attr))
	    {
	      parser->buffer = buffer;
	      break;
	    }

	  tree_add (*attrib, attr);

	  m = g_list_next (m);

	}

      if (m == NULL)
	return TRUE;
      else
	tree_delete (*attrib);

    }

  return FALSE;

}

gpointer rdp_build (Rdp* parser)
{

  gpointer attrib;

  if (rdp_step (parser, parser->start, &attrib))
    return attrib;

  return NULL;

}
